(** Time-stamp: <modified the 01/06/2011 (at 13:43) by Erwan Jahier> *)

(** Here follows a description of the different modules used by this lus2lic compiler.

(1) First of all, the Lustre files are parsed,

   lexer.mll  
   parser.mly
   parserUtils.ml
   lxm.mli/ml

which results into a parse tree containing raw source expressions.

   syntaxTree.ml  -> should rather be called rawSyntaxTab.ml ?

(2) Then, we perform reference checking at module level + model expansion.

   syntaxTab.mli/ml
      syntaxTabUtil.ml/mli
      instanciateModel.mli/ml
      symbolTab.mli/ml (type/const/node)

syntaxTab is a kind of layer above syntaxTree to make things easier afterwards.


(3) Finally, the compilation (type checking+const/type evaluation) is performed.

   compile.ml
      lazyCompiler.mli/ml
        evalConst.mli/ml
        evalType.mli/ml
   compiledData.ml
    

Some misc (eponymous) modules are used along the way.
   errors.ml 
   verbose.mli/ml
   version.ml
   ident.ml
*)


open Verbose
open SyntaxTree
open SyntaxTreeCore
open Lxm
open Errors
open Parsing
open Format

let test_lex ( lexbuf ) = (
  let tk = ref (Lexer.lexer lexbuf) in 
    while !tk <> Parser.TK_EOF do
      match (Lexer.token_code !tk) with 
	  ( co , lxm ) ->
	    Printf.printf "line %3d col %2d to %2d : %15s = \"%s\"\n"
	      (line lxm) (cstart lxm) (cend lxm) co (str lxm) ;
	    tk := (Lexer.lexer lexbuf)
    done
)

(* Retourne un SyntaxTree.t *)
let lus_load lexbuf = 
  let tree = Parser.program Lexer.lexer lexbuf in
    Name.update_fresh_var_prefix ();
    SolveIdent.recognize_predef_op tree
   

type maybe_packed = 
  | Packed of SyntaxTree.pack_or_model
  | Unpacked of SyntaxTree.packbody 



let (get_source_list : string list -> SyntaxTree.pack_or_model list) =
  fun infile_list -> 
    let (get_one_source : string -> string list * maybe_packed list) = 
      fun infile -> 
        let incl_files, l =
          let lexbuf = Global.lexbuf_of_file_name infile in
            if !Global.tlex then test_lex lexbuf;
            match (lus_load lexbuf) with
              | PRPackBody(incl_files, pbdy) -> incl_files, [Unpacked pbdy]
              | PRPack_or_models(incl_files, nsl) -> incl_files, (List.map (fun ns -> Packed ns) nsl)
        in
          (* If included files have a relative path, strange things may happen.
             Hence we make the path absolute, using the directory of the includer.
          *)
        let includer_dir = Filename.dirname infile in
        let fix_dir f = if Filename.is_relative f then Filename.concat includer_dir f else f in
        let incl_files = List.map fix_dir incl_files in
          incl_files, l
    in
    let rec (get_remaining_source_list : maybe_packed list * string list * string list -> 
              maybe_packed list * string list * string list) =
      fun (pack_acc, compiled, to_be_compiled) -> 
        match to_be_compiled with
          | [] -> (pack_acc, compiled, [])
          | infile::tail ->
              let infile = FilenameExtras.simplify infile in
              if List.mem infile compiled then
                get_remaining_source_list (pack_acc, compiled, tail)
              else
                let included_files, pack = get_one_source infile in
                let new_pack_acc = pack_acc@pack in
                  get_remaining_source_list(
                    new_pack_acc, 
                    infile::compiled, 
                    tail@included_files)
    in
    let infile_list = 
      (* We need absolute paths to make sure that files are not
         included several times.  Indeed, otherwise,
         FilenameExtras.simplify may miss some simplifications.  For
         example, consider the files "../../x/toto.lus" and
         "../toto.lus".  They actually refer to the same file if the
         current directory is a sub-directory of "x". Working with
         absolute paths solves the problem.
         
      *)
      let make_it_absolute f = if Filename.is_relative f then Filename.concat (Sys.getcwd ()) f else f in
        List.map make_it_absolute infile_list
    in
    let first_file = assert (infile_list <> []); List.hd infile_list in
    let included_files, first_pack = get_one_source first_file in
    let (pack_list, _compiled_files, included_files) = 
      get_remaining_source_list (first_pack, [first_file], (List.tl infile_list) @ included_files)
    in
    let _ = assert (included_files=[]) in
    let packed_list, unpacked_list = 
      List.fold_left 
        (fun (pl, upl) p -> 
           match p with
             | Packed p ->  p::pl, upl
             | Unpacked up -> pl, up::upl
        )
        ([], [])
        pack_list
    in
    let unpacked_merged_opt = (* All unpacked files are merged into one single package *)
      List.fold_left
        (fun acc pbody -> 
             match acc with
               | None -> Some pbody
               | Some pbody_acc -> 
                   let add tbl x y =
                   (* Let's perform some clashes checks *)
                     if Hashtbl.mem tbl x then
                       let ybis = Hashtbl.find tbl x in
		         print_string ("*** Error: "^(Ident.to_string x)^" is defined twice: \n\t" ^ 
                                         (Lxm.details y.src) ^ "\n\t" ^
                                         (Lxm.details ybis.src) ^ ".\n"); 
                         exit 2
                     else
                       Hashtbl.add tbl x y
                   in
                   Hashtbl.iter (fun x y -> add pbody_acc.pk_const_table x y) pbody.pk_const_table;
                   Hashtbl.iter (fun x y -> add pbody_acc.pk_type_table x y) pbody.pk_type_table;
                   Hashtbl.iter (fun x y -> add pbody_acc.pk_node_table x y) pbody.pk_node_table;
                   Some { pbody_acc with
                            pk_def_list=pbody_acc.pk_def_list@pbody.pk_def_list;
                        }
        )
        None
        unpacked_list
    in
      match unpacked_merged_opt with
        | None -> packed_list
        | Some unpacked_merged ->
            let name = 
              try Filename.chop_extension (Filename.basename first_file) 
              with _ -> 
		print_string ("*** '"^first_file^"': bad file name.\n"); exit 1
            in
            let pi = SyntaxTree.give_pack_this_name (Ident.pack_name_of_string name) unpacked_merged in
            let p = NSPack (Lxm.flagit pi (Lxm.dummy name)) in
              p::packed_list


(*
let dump_entete oc = 
  let time = Unix.localtime (Unix.time ()) in
  let sys_call, _ = Array.fold_left 
    (fun (acc,i) x -> 
       if 70 < i + (String.length x) then 
	 acc ^ "\n--\t\t" ^ x, String.length ("\n--\t\t" ^ x)
       else 
	 acc ^ " " ^ x , (i+1+(String.length x))
    )
    ("",0) 
    Sys.argv
  and 
      date = ( 
	(string_of_int time.Unix.tm_mday) ^ "/" ^
	  (string_of_int (time.Unix.tm_mon+1)) ^  "/" ^
	  (string_of_int (1900+time.Unix.tm_year))
      )
  and time_str = (
    (string_of_int time.Unix.tm_hour) ^  ":" ^
      (if time.Unix.tm_min < 10 then "0" else "") ^
      (string_of_int time.Unix.tm_min) ^   ":" ^
      (if time.Unix.tm_sec < 10 then "0" else "") ^
      (string_of_int time.Unix.tm_sec) 
  )
    (*   and user = Unix.getlogin () *)
  and hostname = Unix.gethostname ()
  in
    output_string oc
      ("-- This file was generated by lus2lic version " ^ Version.str ^ 
         ".\n--\t" ^ sys_call ^ "
-- on " ^ hostname ^ 
         (* 	 "by "^ user ^ *)
         " the " ^ date ^ " at " ^ time_str ^ "\n\n");
    flush oc
*)


let my_exit i =
  close_out !Global.oc;
  if Sys.file_exists !Global.outfile then Sys.remove !Global.outfile;
  exit i

let main = (
  (* Compile.init_appli () ; *)
  (* parse_args (); *)
  let args = MainArgs.parse Sys.argv in
  Verbose.exe ~level:3 (fun () ->
      Gc.set { (Gc.get ()) with Gc.verbose = 0x01 }
  );
  if !Global.run_unit_test then (
    UnifyType.unit_test ();
    exit 0
  );
  if (!Global.infiles = []) then (
    MainArgs.usage stderr args;
    exit 1
  );
  try (
    let nsl = get_source_list !Global.infiles in
    let main_node = 
      if !Global.main_node = "" then None else 
        Some (Ident.idref_of_string !Global.main_node)
    in
      if !Global.outfile <> "" then Global.oc := open_out !Global.outfile;
      (* OBSOLETE *)
      (* dump_entete !Global.oc; *)
      Compile.doit nsl main_node;
      (* OBSOLETE
         LicDump.dump_type_alias !Global.oc;
      *)
      Verbose.exe ~level:3 (fun () ->
         Gc.print_stat stdout
      );
      close_out !Global.oc
  ) with
      Sys_error(s) ->
        prerr_string (s^"\n");
        my_exit 1
    | Global_error s ->
        print_global_error s ;
        my_exit 1
    | Parse_error ->
        print_compile_error (Lxm.last_made ()) "syntax error";
        my_exit 1
    | Unknown_var(lxm,id) -> 
        print_compile_error lxm ("unknown variable (" ^ (Ident.string_of_idref id) ^")")
    | Unknown_constant(lxm,str) -> 
        print_compile_error lxm ("unknown constant (" ^ str ^")")

    | Compile_error(lxm,msg) -> 
        print_compile_error lxm msg ;
        my_exit 1
    | Assert_failure (file, line, col)  -> 
        prerr_string (
          "\n*** oops: an internal error (lus2lic) occurred in file "^ file ^
            ", line " ^ (string_of_int line) ^ ", column " ^
            (string_of_int col) ^ "\n*** when compiling lustre program" ^
            (if List.length !Global.infiles > 1 then "s " else " ") ^
            (String.concat ", " !Global.infiles) ^ "\n"^
            "\n*** You migth want to sent a bug report to jahier@imag.fr\n") ; 
        my_exit 2

        (* | Compile_node_error(nkey,lxm,msg) -> ( *)
        (* print_compile_node_error nkey lxm msg ; *)
        (* exit 1 *)
        (* ) *)
        (* | Global_node_error(nkey,msg) -> ( *)
        (* print_global_node_error nkey msg ; *)
        (* exit 1 *)
        (* ) *)

)
